
# THIS FILE IS GENERATED FROM HIFIVE SETUP.PY
short_version = '1.3.0'
version = '1.3.0'
full_version = '1.3.0'
git_revision = '1194212c4d006c5a05222384d2b9ed842ced4739'
release = True

if not release:
    version = full_version
